from enum import Enum
from typing import Any, cast

from starlite.exceptions import ImproperlyConfiguredException


class HttpMethod(str, Enum):
    GET = "get"
    POST = "post"
    PUT = "put"
    PATCH = "patch"
    DELETE = "delete"

    @classmethod
    def is_http_method(cls, value: Any) -> bool:
        """Validates that a given value is a member of the HttpMethod enum"""
        return isinstance(value, str) and value.lower() in list(cls)

    @classmethod
    def from_str(cls, value: Any) -> "HttpMethod":
        """Given a string value, return an enum member or raise a ValueError"""
        if cls.is_http_method(value):
            return cast(HttpMethod, value.lower())
        raise ImproperlyConfiguredException(f"value {value} is not a supported http method")


class MediaType(str, Enum):
    JSON = "application/json"
    HTML = "text/html"
    TEXT = "text/plain"


class OpenAPIMediaType(str, Enum):
    OPENAPI_YAML = "application/vnd.oai.openapi"
    OPENAPI_JSON = "application/vnd.oai.openapi+json"


class RequestEncodingType(str, Enum):
    JSON = "application/json"
    MULTI_PART = "multipart/form-data"
    URL_ENCODED = "application/x-www-form-urlencoded"



# flake8: noqa
from .model import (
    convert_dataclass_to_model,
    create_function_signature_model,
    create_parsed_model_field,
)
from .sequence import find_index, unique
from .url import join_paths, normalize_path

__all__ = [
    "join_paths",
    "normalize_path",
    "find_index",
    "unique",
    "create_parsed_model_field",
    "create_function_signature_model",
    "convert_dataclass_to_model",
]



from inspect import Signature, getfullargspec, isclass
from typing import Any, Dict, Tuple, Type, cast

from pydantic import BaseConfig, BaseModel, create_model
from pydantic.fields import ModelField
from pydantic.typing import AnyCallable
from pydantic_factories import ModelFactory
from pydantic_factories.utils import create_model_from_dataclass

from starlite.exceptions import ImproperlyConfiguredException


def create_function_signature_model(fn: AnyCallable) -> Type[BaseModel]:
    """
    Creates a pydantic model for the signature of a given function
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    try:
        signature = Signature.from_callable(fn)
        field_definitions: Dict[str, Tuple[Any, Any]] = {}
        for key, value in getfullargspec(fn).annotations.items():
            if key == "return":
                continue
            parameter = signature.parameters[key]
            if key == "request":
                field_definitions[key] = (Any, ...)
            elif ModelFactory.is_constrained_field(parameter.default):
                field_definitions[key] = (parameter.default, ...)
            elif parameter.default is not signature.empty:
                field_definitions[key] = (value, parameter.default)
            elif not repr(parameter.annotation).startswith("typing.Optional"):
                field_definitions[key] = (value, ...)
            else:
                field_definitions[key] = (value, None)
        name = (fn.__name__ if hasattr(fn, "__name__") else "anonymous") + "SignatureModel"
        return create_model(name, __config__=Config, **field_definitions)  # type: ignore
    except TypeError as e:
        raise ImproperlyConfiguredException(repr(e)) from e


def create_parsed_model_field(value: Type[Any]) -> ModelField:
    """Create a pydantic model with the passed in value as its sole field, and return the parsed field"""

    model = create_model("temp", **{"value": (value, ... if not repr(value).startswith("typing.Optional") else None)})  # type: ignore
    return cast(BaseModel, model).__fields__["value"]


_dataclass_model_map: Dict[Any, Type[BaseModel]] = {}


def convert_dataclass_to_model(dataclass: Any) -> Type[BaseModel]:
    """Converts a dataclass to a pydantic model and memoizes the result"""
    if not isclass(dataclass) and hasattr(dataclass, "__class__"):
        dataclass = dataclass.__class__
    if not _dataclass_model_map.get(dataclass):
        _dataclass_model_map[dataclass] = create_model_from_dataclass(dataclass)
    return _dataclass_model_map[dataclass]



from typing import TYPE_CHECKING, cast

from openapi_schema_pydantic import Operation, PathItem
from pydantic.typing import AnyCallable
from starlette.routing import get_name

from starlite.openapi.parameters import create_parameters
from starlite.openapi.request_body import create_request_body
from starlite.openapi.responses import create_responses
from starlite.utils.model import create_function_signature_model

if TYPE_CHECKING:  # pragma: no cover
    from starlite.routing import Route


def create_path_item(route: "Route", create_examples: bool) -> PathItem:
    """
    Create a PathItem model for the given route parsing all http_methods into Operation Models
    """
    path_item = PathItem()
    for http_method, route_handler in route.route_handler_map.items():
        if route_handler.include_in_schema:
            route_handler_fn = cast(AnyCallable, route_handler.fn)
            handler_fields = create_function_signature_model(fn=route_handler_fn).__fields__
            parameters = (
                create_parameters(
                    route_handler=route_handler,
                    handler_fields=handler_fields,
                    path_parameters=route.path_parameters,
                    generate_examples=create_examples,
                )
                or None
            )
            raises_validation_error = bool("data" in handler_fields or path_item.parameters or parameters)
            handler_name = get_name(route_handler_fn)
            request_body = None
            if "data" in handler_fields:
                request_body = create_request_body(field=handler_fields["data"], generate_examples=create_examples)
            operation = Operation(
                operationId=route_handler.operation_id or handler_name,
                tags=route_handler.tags,
                summary=route_handler.summary,
                description=route_handler.description,
                deprecated=route_handler.deprecated,
                responses=create_responses(
                    route_handler=route_handler,
                    raises_validation_error=raises_validation_error,
                    generate_examples=create_examples,
                ),
                requestBody=request_body,
                parameters=parameters,
            )
            setattr(path_item, http_method, operation)
    return path_item



from enum import Enum


class OpenAPIFormat(str, Enum):
    """
    Formats extracted from:
    https://datatracker.ietf.org/doc/html/draft-bhutton-json-schema-validation-00#page-13
    """

    DATE = "date"
    DATE_TIME = "date-time"
    TIME = "time"
    DURATION = "duration"
    URL = "url"
    EMAIL = "email"
    IDN_EMAIL = "idn-email"
    HOST_NAME = "hostname"
    IDN_HOST_NAME = "idn-hostname"
    IPV4 = "ipv4"
    IPV6 = "ipv6"
    URI = "uri"
    URI_REFERENCE = "uri-reference"
    URI_TEMPLATE = "uri-template"
    JSON_POINTER = "json-pointer"
    RELATIVE_JSON_POINTER = "relative-json-pointer"
    IRI = "iri-reference"
    IRI_REFERENCE = "iri-reference"
    UUID = "uuid"
    REGEX = "regex"


class OpenAPIType(str, Enum):
    ARRAY = "array"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    NULL = "null"
    NUMBER = "number"
    OBJECT = "object"
    STRING = "string"



from typing import Any, Dict, List
from uuid import UUID

from openapi_schema_pydantic import Parameter, Schema
from pydantic.fields import ModelField
from typing_extensions import Type

from starlite.handlers import RouteHandler
from starlite.openapi.schema import create_schema


def create_path_parameter_schema(path_parameter: str, field: ModelField, generate_examples: bool) -> Schema:
    """Create a path parameter from the given path_param string in the format param_name:type"""
    param_type_map: Dict[str, Type[Any]] = {
        "str": str,
        "float": float,
        "int": int,
        "uuid": UUID,
    }
    parameter_type = path_parameter.split(":")[1]
    assert parameter_type in param_type_map, f"Unsupported path param type {parameter_type}"
    field.sub_fields = None
    field.outer_type_ = param_type_map[parameter_type]
    return create_schema(field=field, generate_examples=generate_examples)


def create_parameters(
    route_handler: RouteHandler,
    handler_fields: Dict[str, ModelField],
    path_parameters: List[str],
    generate_examples: bool,
) -> List[Parameter]:
    """
    Create a list of path/query/header Parameter models for the given PathHandler
    """
    path_parameter_names = [path_param.split(":")[0] for path_param in path_parameters]
    parameters: List[Parameter] = []
    ignored_fields = [
        "data",
        "request",
        "headers",
        *list(route_handler.resolve_dependencies().keys()),
    ]
    for f_name, field in handler_fields.items():
        if f_name not in ignored_fields:
            schema = None
            param_in = "query"
            required = field.required
            extra = field.field_info.extra
            header_key = extra.get("header")
            cookie_key = extra.get("cookie")
            query_key = extra.get("query")
            if f_name in path_parameter_names:
                param_in = "path"
                required = True
                schema = create_path_parameter_schema(
                    path_parameter=[p for p in path_parameters if f_name in p][0],
                    field=field,
                    generate_examples=generate_examples,
                )
            elif header_key:
                f_name = header_key
                param_in = "header"
                required = field.field_info.extra["required"]
            elif cookie_key:
                f_name = cookie_key
                param_in = "cookie"
                required = field.field_info.extra["required"]
            elif query_key:
                f_name = query_key
                param_in = "query"
                required = field.required
            if not schema:
                schema = create_schema(field=field, generate_examples=generate_examples)
            parameters.append(
                Parameter(
                    name=f_name,
                    param_in=param_in,
                    required=required,
                    param_schema=schema,
                    description=schema.description,
                )
            )
    return parameters



import re

CAPITAL_LETTERS_PATTERN = re.compile(r"(?=[A-Z])")


def pascal_case_to_text(s: str) -> str:
    """Given a 'PascalCased' string, return its split form- 'Pascal Cased'"""
    return " ".join(re.split(CAPITAL_LETTERS_PATTERN, s)).strip()



from enum import Enum
from inspect import isawaitable
from typing import TYPE_CHECKING, Any, Dict, Generic, List, TypeVar, Union, cast

from orjson import loads
from pydantic.error_wrappers import ValidationError, display_errors
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON, ModelField
from pydantic.typing import AnyCallable
from starlette.datastructures import UploadFile
from starlette.requests import Request as StarletteRequest
from starlette.responses import FileResponse, RedirectResponse
from starlette.responses import Response as StarletteResponse
from starlette.responses import StreamingResponse

from starlite.controller import Controller
from starlite.enums import HttpMethod, RequestEncodingType
from starlite.exceptions import ImproperlyConfiguredException, ValidationException
from starlite.types import File, Redirect, Stream

if TYPE_CHECKING:  # pragma: no cover
    from starlite.app import Starlite
    from starlite.handlers import RouteHandler

User = TypeVar("User")
Auth = TypeVar("Auth")


class Request(StarletteRequest, Generic[User, Auth]):
    @property
    def app(self) -> "Starlite":
        return cast("Starlite", self.scope["app"])

    @property
    def user(self) -> User:
        assert "user" in self.scope, "user is not defined in scope, you should install an AuthMiddleware to set it"
        return cast(User, self.scope["user"])

    @property
    def auth(self) -> Auth:
        assert "auth" in self.scope, "auth is not defined in scope, you should install an AuthMiddleware to set it"
        return cast(Auth, self.scope["auth"])


def parse_query_params(request: Request) -> Dict[str, Any]:
    """
    Parses and normalize a given request's query parameters into a regular dictionary

    supports list query params
    """
    params: Dict[str, Union[str, List[str]]] = {}
    try:
        for key, value in request.query_params.multi_items():
            if value in ["True", "true"]:
                value = True  # type: ignore
            elif value in ["False", "false"]:
                value = False  # type: ignore
            param = params.get(key)
            if param:
                if isinstance(param, str):
                    params[key] = [param, value]
                else:
                    params[key] = [*cast(List[Any], param), value]
            else:
                params[key] = value
        return params
    except KeyError:
        return params


async def get_request_data(request: Request, field: ModelField) -> Any:
    """Given a request, parse its data - either as json or form data and return it"""
    if request.method.lower() == HttpMethod.GET:
        raise ImproperlyConfiguredException("'data' kwarg is unsupported for GET requests")
    media_type = field.field_info.extra.get("media_type")
    if not media_type or media_type == RequestEncodingType.JSON:
        body = await request.body()
        json_data = request._json = loads(body)
        return json_data
    form_data = await request.form()
    as_dict = dict(form_data.multi_items())
    if media_type == RequestEncodingType.MULTI_PART:
        if field.shape is SHAPE_LIST:
            return list(as_dict.values())
        if field.shape is SHAPE_SINGLETON and field.type_ is UploadFile and as_dict:
            return list(as_dict.values())[0]
    return as_dict


def get_request_parameters(
    request: Request,
    field_name: str,
    field: ModelField,
    query_params: Dict[str, Any],
    header_params: Dict[str, Any],
) -> Any:
    """Extract path, query, header and cookie parameters correlating to field_names from the request"""
    if field_name in request.path_params:
        return request.path_params[field_name]
    if field_name in query_params:
        return query_params[field_name]

    extra = field.field_info.extra
    parameter_name = None
    source = None

    if extra.get("query"):
        parameter_name = extra["query"]
        source = query_params
    if extra.get("header"):
        parameter_name = extra["header"]
        source = header_params
    if extra.get("cookie"):
        parameter_name = extra["cookie"]
        source = request.cookies
    if parameter_name and source:
        parameter_is_required = extra["required"]
        try:
            return source[parameter_name]
        except KeyError as e:
            if parameter_is_required:
                raise ValidationException(f"Missing required parameter {parameter_name}") from e
    return None


async def get_model_kwargs_from_request(request: Request, fields: Dict[str, ModelField]) -> Dict[str, Any]:
    """
    Given a function's signature Model fields, populate its kwargs from the Request object
    """
    kwargs: Dict[str, Any] = {}
    query_params = parse_query_params(request=request)
    header_params = dict(request.headers.items())
    for field_name, field in fields.items():
        if field_name == "request":
            kwargs["request"] = request
        elif field_name == "headers":
            kwargs["headers"] = header_params
        elif field_name == "cookies":
            kwargs["cookies"] = request.cookies
        elif field_name == "query":
            kwargs["query"] = query_params
        elif field_name == "data":
            kwargs["data"] = await get_request_data(request=request, field=field)
        else:
            kwargs[field_name] = get_request_parameters(
                request=request,
                field_name=field_name,
                field=field,
                query_params=query_params,
                header_params=header_params,
            )
    return kwargs


async def get_http_handler_parameters(route_handler: "RouteHandler", request: Request) -> Dict[str, Any]:
    """
    Parse a given http handler function and return values matching function parameter keys
    """
    model = route_handler.signature_model
    assert model, "route handler has no signature model"
    try:
        # dependency injection
        dependencies: Dict[str, Any] = {}
        for key, provider in route_handler.resolve_dependencies().items():
            provider_kwargs = await get_model_kwargs_from_request(
                request=request, fields=provider.signature_model.__fields__
            )
            value = provider(**provider.signature_model(**provider_kwargs).dict())
            if isawaitable(value):
                value = await value
            dependencies[key] = value
        model_kwargs = await get_model_kwargs_from_request(
            request=request, fields={k: v for k, v in model.__fields__.items() if k not in dependencies}
        )
        # we return the model's attributes as a dict in order to preserve any nested models
        fields = list(model.__fields__.keys())
        return {key: model(**model_kwargs, **dependencies).__getattribute__(key) for key in fields}
    except ValidationError as e:
        raise ValidationException(
            detail=f"Validation failed for {request.method} {request.url}:\n\n{display_errors(e.errors())}"
        ) from e


async def handle_request(route_handler: "RouteHandler", request: Request) -> StarletteResponse:
    """
    Handles a given request by both calling the passed in function,
    and parsing the RouteHandler stored as an attribute on it.
    """
    params = await get_http_handler_parameters(route_handler=route_handler, request=request)

    for guard in route_handler.resolve_guards():
        result = guard(request, route_handler.copy())
        if isawaitable(result):
            await result

    endpoint = cast(AnyCallable, route_handler.fn)

    if isinstance(route_handler.owner, Controller):
        data = endpoint(route_handler.owner, **params)
    else:
        data = endpoint(**params)
    if isawaitable(data):
        data = await data
    if isinstance(data, StarletteResponse):
        return data

    status_code = cast(int, route_handler.status_code)
    headers = {k: v.value for k, v in route_handler.resolve_response_headers().items()}
    if isinstance(data, Redirect):
        return RedirectResponse(headers=headers, status_code=status_code, url=data.path)

    media_type = (
        route_handler.media_type.value if isinstance(route_handler.media_type, Enum) else route_handler.media_type
    )
    if isinstance(data, File):
        return FileResponse(media_type=media_type, headers=headers, **data.dict())
    if isinstance(data, Stream):
        return StreamingResponse(content=data.iterator, status_code=status_code, media_type=media_type, headers=headers)

    response_class = route_handler.resolve_response_class()
    return response_class(
        headers=headers,
        status_code=status_code,
        content=data,
        media_type=media_type,
    )



import re
from inspect import isclass
from typing import Any, Dict, List, Optional, Union, cast

from pydantic import validate_arguments
from pydantic.typing import AnyCallable
from starlette.routing import Route as StarletteRoute
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import Type

from starlite.controller import Controller
from starlite.enums import HttpMethod
from starlite.exceptions import ImproperlyConfiguredException
from starlite.handlers import RouteHandler
from starlite.provide import Provide
from starlite.request import Request, handle_request
from starlite.response import Response
from starlite.types import Guard, ResponseHeader
from starlite.utils.sequence import find_index, unique
from starlite.utils.url import join_paths, normalize_path

param_match_regex = re.compile(r"{(.*?)}")


class Route(StarletteRoute):
    route_handler_map: Dict[HttpMethod, RouteHandler]

    @validate_arguments()
    def __init__(
        self,
        *,
        path: str,
        route_handlers: Union[RouteHandler, List[RouteHandler]],
    ):
        self.route_handler_map = self.parse_route_handlers(
            route_handlers=route_handlers if isinstance(route_handlers, list) else [route_handlers], path=path
        )
        # we are passing a dud lambda function as the endpoint kwarg here because we are setting self.app ourselves
        super().__init__(path=path, methods=[method.upper() for method in self.route_handler_map], endpoint=lambda x: x)
        self.app = self.create_endpoint_handler(self.route_handler_map)
        self.path_parameters: List[str] = param_match_regex.findall(self.path)
        for parameter in self.path_parameters:
            if ":" not in parameter or not parameter.split(":")[1]:
                raise ImproperlyConfiguredException("path parameter must declare a type: '{parameter_name:type}'")

    @staticmethod  #
    def parse_route_handlers(route_handlers: List[RouteHandler], path: str) -> Dict[HttpMethod, RouteHandler]:
        """
        Parses the passed in route_handlers and returns a mapping of http-methods and route handlers
        """
        mapped_route_handlers: Dict[HttpMethod, RouteHandler] = {}
        for route_handler in route_handlers:
            for http_method in route_handler.http_methods:
                if mapped_route_handlers.get(http_method):
                    raise ImproperlyConfiguredException(
                        f"handler already registered for path {path!r} and http method {http_method}"
                    )
                mapped_route_handlers[http_method] = route_handler
        return mapped_route_handlers

    @staticmethod
    def create_endpoint_handler(http_handler_mapping: Dict[HttpMethod, RouteHandler]) -> ASGIApp:
        """
        Create an ASGIApp that routes to the correct route handler based on the scope.

        Using this method, Starlite is able to support different handler functions for the same path.
        """

        async def endpoint_handler(scope: Scope, receive: Receive, send: Send) -> None:
            request: Request[Any, Any] = Request(scope, receive=receive, send=send)
            request_method = HttpMethod.from_str(request.method)
            handler = http_handler_mapping[request_method]
            response = await handle_request(route_handler=handler, request=request)
            await response(scope, receive, send)

        return endpoint_handler


# noinspection PyMethodOverriding
class Router:
    def __init__(
        self,
        *,
        path: str,
        route_handlers: List[Union[Type[Controller], RouteHandler, "Router", AnyCallable]],
        dependencies: Optional[Dict[str, Provide]] = None,
        guards: Optional[List[Guard]] = None,
        response_class: Optional[Type[Response]] = None,
        response_headers: Optional[Dict[str, ResponseHeader]] = None,
    ):
        self.owner: Optional["Router"] = None
        self.routes: List[Route] = []
        self.path = normalize_path(path)
        self.response_class = response_class
        self.dependencies = dependencies
        self.response_headers = response_headers
        self.guards = guards
        for route_handler in route_handlers or []:
            self.register(value=route_handler)

    @property
    def route_handler_method_map(self) -> Dict[str, Dict[HttpMethod, RouteHandler]]:
        """
        Returns dictionary that maps paths (keys) to a list of route handler functions (values)
        """
        r_map: Dict[str, Dict[HttpMethod, RouteHandler]] = {}
        for r in self.routes:
            if not r_map.get(r.path):
                r_map[r.path] = {}
            for method, handler in r.route_handler_map.items():
                r_map[r.path][method] = handler
        return r_map

    @staticmethod
    def create_handler_http_method_map(
        value: Union[Controller, RouteHandler, "Router"],
    ) -> Dict[str, Dict[HttpMethod, RouteHandler]]:
        """
        Maps route handlers to http methods
        """
        handlers_map: Dict[str, Dict[HttpMethod, RouteHandler]] = {}
        if isinstance(value, RouteHandler):
            handlers_map[value.path or ""] = {http_method: value for http_method in value.http_methods}
        elif isinstance(value, Router):
            handlers_map = value.route_handler_method_map
        else:
            # we reassign the variable to give it a clearer meaning
            for route_handler in value.get_route_handlers():
                path = join_paths([value.path, route_handler.path]) if route_handler.path else value.path
                if not handlers_map.get(path):
                    handlers_map[path] = {}
                for http_method in route_handler.http_methods:
                    handlers_map[path][http_method] = route_handler
        return handlers_map

    def validate_registration_value(
        self, value: Union[Type[Controller], RouteHandler, "Router", AnyCallable]
    ) -> Union[Controller, RouteHandler, "Router"]:
        """
        Validates that the value passed to the register method is supported
        """
        if isclass(value) and issubclass(cast(Type[Controller], value), Controller):
            return cast(Type[Controller], value)(owner=self)
        if not isinstance(value, (Router, RouteHandler)):
            raise ImproperlyConfiguredException(
                "Unsupported value passed to `Router.register`. "
                "If you passed in a function or method, "
                "make sure to decorate it first with one of the routing decorators"
            )
        if isinstance(value, Router):
            if value.owner:
                raise ImproperlyConfiguredException(f"Router with path {value.path} has already been registered")
            if value is self:
                raise ImproperlyConfiguredException("Cannot register a router on itself")
            if not value.owner:
                value.owner = self
        else:
            # the route handler is copied to ensure each time the route handler is registerd,
            # we get an instance with a unique owner
            value = value.copy()
            value.owner = self
        return cast(Union[Controller, RouteHandler, "Router"], value)

    def register(self, value: Union[Type[Controller], RouteHandler, "Router", AnyCallable]) -> None:
        """
        Register a Controller, Route instance or RouteHandler on the router

        Accepts a subclass or instance of Controller, an instance of Router or a function/method that has been decorated
        by any of the routing decorators (e.g. route, get, post...) exported from 'starlite.routing'
        """
        validated_value = self.validate_registration_value(value)
        if not validated_value.owner:
            validated_value.owner = self
        handlers_map = self.create_handler_http_method_map(value=validated_value)

        for route_path, method_map in handlers_map.items():
            path = join_paths([self.path, route_path])
            route_handlers = unique(method_map.values())
            if self.route_handler_method_map.get(path):
                existing_route_index = find_index(
                    self.routes, lambda x: x.path == path  # pylint: disable=cell-var-from-loop
                )
                assert existing_route_index != -1, "unable to find_index existing route index"
                self.routes[existing_route_index] = Route(
                    path=path,
                    route_handlers=unique([*list(self.route_handler_method_map[path].values()), *route_handlers]),
                )
            else:
                self.routes.append(Route(path=path, route_handlers=route_handlers))



from typing import Dict, List, Optional, Union

from openapi_schema_pydantic import OpenAPI
from openapi_schema_pydantic.util import construct_open_api_with_schema_class
from pydantic import validate_arguments
from pydantic.typing import AnyCallable, NoArgAnyCallable
from starlette.datastructures import State
from starlette.exceptions import ExceptionMiddleware
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.routing import Router as StarletteRouter
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR
from starlette.types import ASGIApp, Receive, Scope, Send
from typing_extensions import Type

from starlite.config import CORSConfig, OpenAPIConfig
from starlite.controller import Controller
from starlite.enums import MediaType
from starlite.exceptions import HTTPException
from starlite.handlers import RouteHandler
from starlite.openapi.path_item import create_path_item
from starlite.provide import Provide
from starlite.request import Request
from starlite.response import Response
from starlite.routing import Router
from starlite.types import ExceptionHandler, Guard, MiddlewareProtocol, ResponseHeader


class Starlite(Router):
    @validate_arguments(config={"arbitrary_types_allowed": True})
    def __init__(
        self,
        *,
        allowed_hosts: Optional[List[str]] = None,
        cors_config: Optional[CORSConfig] = None,
        debug: bool = False,
        dependencies: Optional[Dict[str, Provide]] = None,
        exception_handlers: Optional[Dict[Union[int, Type[Exception]], ExceptionHandler]] = None,
        guards: Optional[List[Guard]] = None,
        middleware: Optional[List[Union[Middleware, Type[BaseHTTPMiddleware], Type[MiddlewareProtocol]]]] = None,
        on_shutdown: Optional[List[NoArgAnyCallable]] = None,
        on_startup: Optional[List[NoArgAnyCallable]] = None,
        openapi_config: Optional[OpenAPIConfig] = None,
        redirect_slashes: bool = True,
        response_class: Optional[Type[Response]] = None,
        response_headers: Optional[Dict[str, ResponseHeader]] = None,
        route_handlers: List[Union[Type[Controller], RouteHandler, Router, AnyCallable]],
    ):
        self.debug = debug
        self.state = State()
        super().__init__(
            dependencies=dependencies,
            guards=guards,
            path="",
            response_class=response_class,
            response_headers=response_headers,
            route_handlers=route_handlers,
        )
        self.asgi_router: StarletteRouter = StarletteRouter(
            redirect_slashes=redirect_slashes,
            on_shutdown=on_shutdown or [],
            on_startup=on_startup or [],
            routes=self.routes,
        )
        self.exception_handlers: Dict[Union[int, Type[Exception]], ExceptionHandler] = {
            StarletteHTTPException: self.handle_http_exception,
            **(exception_handlers or {}),
        }
        self.middleware_stack: ASGIApp = self.build_middleware_stack(
            user_middleware=middleware or [], cors_config=cors_config, allowed_hosts=allowed_hosts
        )
        if openapi_config:
            self.openapi_schema = self.create_openapi_schema_model(openapi_config=openapi_config)
        else:
            self.openapi_schema = None

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self
        await self.middleware_stack(scope, receive, send)

    def build_middleware_stack(
        self,
        user_middleware: List[Union[Middleware, Type[BaseHTTPMiddleware], Type[MiddlewareProtocol]]],
        cors_config: Optional[CORSConfig],
        allowed_hosts: Optional[List[str]],
    ) -> ASGIApp:
        """
        Builds the middleware by sandwiching the user middleware between
        the Starlette ExceptionMiddleware and the starlette ServerErrorMiddleware
        """
        current_app: ASGIApp = ExceptionMiddleware(
            app=self.asgi_router, handlers=self.exception_handlers, debug=self.debug
        )
        if allowed_hosts:
            current_app = TrustedHostMiddleware(app=current_app, allowed_hosts=allowed_hosts)
        if cors_config:
            current_app = CORSMiddleware(app=current_app, **cors_config.dict())

        for middleware in user_middleware:
            if isinstance(middleware, Middleware):
                current_app = middleware.cls(app=current_app, **middleware.options)
            else:
                current_app = middleware(app=current_app)
        return ServerErrorMiddleware(
            handler=self.exception_handlers.get(HTTP_500_INTERNAL_SERVER_ERROR, self.handle_http_exception),
            debug=self.debug,
            app=current_app,
        )

    @staticmethod
    def handle_http_exception(_: Request, exc: Exception) -> Response:
        """Default handler for exceptions subclassed from HTTPException"""
        status_code = exc.status_code if isinstance(exc, StarletteHTTPException) else HTTP_500_INTERNAL_SERVER_ERROR
        if isinstance(exc, HTTPException):
            content = {"detail": exc.detail, "extra": exc.extra}
        elif isinstance(exc, StarletteHTTPException):
            content = {"detail": exc.detail}
        else:
            content = {"detail": repr(exc)}
        return Response(
            media_type=MediaType.JSON,
            content=content,
            status_code=status_code,
        )

    def create_openapi_schema_model(self, openapi_config: OpenAPIConfig) -> OpenAPI:
        """
        Updates the OpenAPI schema with all paths registered on the root router
        """
        openapi_schema = openapi_config.to_openapi_schema()
        openapi_schema.paths = {}
        for route in self.routes:
            if route.include_in_schema and (route.path_format or "/") not in openapi_schema.paths:
                openapi_schema.paths[route.path_format or "/"] = create_path_item(
                    route=route, create_examples=openapi_config.create_examples
                )
        return construct_open_api_with_schema_class(openapi_schema)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starlite', 'starlite.openapi', 'starlite.utils']

package_data = \
{'': ['*']}

install_requires = \
['openapi-schema-pydantic',
 'orjson',
 'pydantic',
 'pydantic-factories',
 'python-multipart',
 'pyyaml',
 'requests',
 'starlette',
 'typing-extensions']

setup_kwargs = {
    'name': 'starlite',
    'version': '0.1.2',
    'description': 'Light-weight and flexible ASGI API Framework',
    'long_description': '<img alt="Starlite logo" src="./starlite-logo.svg" width=100%, height="auto">\n\n<div align="center">\n\n![PyPI - License](https://img.shields.io/pypi/l/starlite?color=blue)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/starlite)\n\n[![Coverage](https://sonarcloud.io/api/project_badges/measure?project=Goldziher_starlite&metric=coverage)](https://sonarcloud.io/summary/new_code?id=Goldziher_starlite)\n[![Vulnerabilities](https://sonarcloud.io/api/project_badges/measure?project=Goldziher_starlite&metric=vulnerabilities)](https://sonarcloud.io/summary/new_code?id=Goldziher_starlite)\n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=Goldziher_starlite&metric=alert_status)](https://sonarcloud.io/summary/new_code?id=Goldziher_starlite)\n[![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=Goldziher_starlite&metric=sqale_rating)](https://sonarcloud.io/summary/new_code?id=Goldziher_starlite)\n[![Reliability Rating](https://sonarcloud.io/api/project_badges/measure?project=Goldziher_starlite&metric=reliability_rating)](https://sonarcloud.io/summary/new_code?id=Goldziher_starlite)\n[![Security Rating](https://sonarcloud.io/api/project_badges/measure?project=Goldziher_starlite&metric=security_rating)](https://sonarcloud.io/summary/new_code?id=Goldziher_starlite)\n\n[![Discord](https://img.shields.io/discord/919193495116337154?color=blue&label=chat%20on%20discord&logo=discord)](https://discord.gg/X3FJqy8d2j)\n</div>\n\n# Starlite\n\nStarlite is a light, opinionated and flexible ASGI API framework built on top\nof [pydantic](https://github.com/samuelcolvin/pydantic) and [Starlette](https://github.com/encode/starlette).\n\nCheck out the [Starlite documentation 📚](https://goldziher.github.io/starlite/)\n\n## Core Features\n\n* 👉 OpenAPI 3.1 schema generation\n* 👉 built-in [Redoc](https://github.com/Redocly/redoc) based OpenAPI UI\n* 👉 class based controllers\n* 👉 decorators based configuration\n* 👉 extended testing support\n* 👉 extensive typing support including inference, validation and parsing\n* 👉 full async (ASGI) support\n* 👉 layered dependency injection\n* 👉 route guards based authorization\n* 👉 simple middleware and authentication\n* 👉 support for pydantic models and pydantic dataclasses\n* 👉 support for standard library dataclasses\n* 👉 ultra-fast json serialization and deserialization using [orjson](https://github.com/ijl/orjson)\n\n## Installation\n\nUsing your package manager of choice:\n\n```shell\npip install starlite\n```\n\nOR\n\n```sh\npoetry add starlite\n```\n\nOR\n\n```sh\npipenv install starlite\n```\n\n## Minimal Example\n\nDefine your data model using pydantic or any library based on it (see for example ormar, beanie, SQLModel etc.):\n\n```python title="my_app/models/user.py"\nfrom pydantic import BaseModel, UUID4\n\n\nclass User(BaseModel):\n    first_name: str\n    last_name: str\n    id: UUID4\n```\n\nYou can alternatively use a dataclass, either the standard library one or the one from pydantic:\n\n```python title="my_app/models/user.py"\nfrom uuid import UUID\n\n# from pydantic.dataclasses import dataclass\nfrom dataclasses import dataclass\n\n@dataclass\nclass User:\n    first_name: str\n    last_name: str\n    id: UUID\n```\n\nDefine a Controller for your data model:\n\n```python title="my_app/controllers/user.py"\nfrom pydantic import UUID4\nfrom starlite.controller import Controller\nfrom starlite.handlers import get, post, put, patch, delete\nfrom starlite.types import Partial\n\nfrom my_app.models import User\n\n\nclass UserController(Controller):\n    path = "/users"\n\n    @post()\n    async def create(self, data: User) -> User:\n        ...\n\n    @get()\n    async def get_users(self) -> list[User]:\n        ...\n\n    @patch(path="/{user_id:uuid}")\n    async def partial_update_user(self, user_id: UUID4, data: Partial[User]) -> User:\n        ...\n\n    @put(path="/{user_id:uuid}")\n    async def update_user(self, user_id: UUID4, data: list[User]) -> list[User]:\n        ...\n\n    @get(path="/{user_id:uuid}")\n    async def get_user_by_id(self, user_id: UUID4) -> User:\n        ...\n\n    @delete(path="/{user_id:uuid}")\n    async def delete_user_by_id(self, user_id: UUID4) -> User:\n        ...\n\n```\n\nImport your controller into your application\'s entry-point and pass it to Starlite when instantiating your app:\n\n```python title="my_app/main.py"\nfrom starlite import Starlite\n\nfrom my_app.controllers.user import UserController\n\napp = Starlite(route_handlers=[UserController])\n```\n\nTo run you application, use an ASGI server such as [uvicorn](https://www.uvicorn.org/):\n\n```shell\nuvicorn my_app.main:app --reload\n```\n\n### Contributing\n\nStarlite is open to contributions big and small. You can always [join our discord](https://discord.gg/X3FJqy8d2j) server\nto discuss contributions and project maintenance. For guidelines on how to contribute, please\nsee [the contribution guide](CONTRIBUTING.md).\n',
    'author': "Na'aman Hirschfeld",
    'author_email': 'Naaman.Hirschfeld@sprylab.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Goldziher/starlite',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# flake8: noqa
from .app import Starlite
from .config import CORSConfig, OpenAPIConfig
from .controller import Controller
from .enums import HttpMethod, MediaType, OpenAPIMediaType, RequestEncodingType
from .exceptions import (
    HTTPException,
    ImproperlyConfiguredException,
    InternalServerException,
    NotAuthorizedException,
    PermissionDeniedException,
    ServiceUnavailableException,
    StarLiteException,
)
from .handlers import RouteHandler, delete, get, patch, post, put, route
from .logging import LoggingConfig
from .middleware import AbstractAuthenticationMiddleware, AuthenticationResult
from .openapi.controller import OpenAPIController
from .params import Body, Parameter
from .provide import Provide
from .request import Request
from .response import Response
from .routing import Route, Router
from .testing import TestClient, create_test_client, create_test_request
from .types import File, MiddlewareProtocol, Partial, Redirect, ResponseHeader, Stream

__all__ = [
    "AbstractAuthenticationMiddleware",
    "AuthenticationResult",
    "Body",
    "CORSConfig",
    "Controller",
    "File",
    "HTTPException",
    "HttpMethod",
    "ImproperlyConfiguredException",
    "InternalServerException",
    "LoggingConfig",
    "MediaType",
    "MiddlewareProtocol",
    "NotAuthorizedException",
    "OpenAPIConfig",
    "OpenAPIController",
    "OpenAPIMediaType",
    "Parameter",
    "Partial",
    "PermissionDeniedException",
    "Provide",
    "Redirect",
    "Request",
    "RequestEncodingType",
    "Response",
    "ResponseHeader",
    "Route",
    "RouteHandler",
    "Router",
    "ServiceUnavailableException",
    "StarLiteException",
    "Starlite",
    "Stream",
    "TestClient",
    "create_test_client",
    "create_test_request",
    "delete",
    "get",
    "patch",
    "post",
    "put",
    "route",
]
